with Ada.Strings.Unbounded.Text_IO, Ints, Ints.IO, Messages;
use Ints, Ints.IO, Messages;

pragma Optimize (Off);

package body Balls is

  RepZero:    constant Rep  := Val(Zero);
  RepOne:     constant Rep  := Val(One);
  RepThree:   constant Rep  := Val(Three);
  RepNegOne:  constant Rep  := Val(NegOne);
  RepEighth:  constant Rep  := RepOne/Val(8);
  RepTiny:    constant Rep  := Val(Tiny);
  BallZero:   constant Ball := (RepZero,RepZero);
  BallOne:    constant Ball := (RepOne,RepZero);

  function "="(S1,S2: Ball) return Boolean is
--- different from other comparisons
  begin
    return (S1.C=S2.C) and then (S1.R=S2.R);
  end "=";

  function IsNumeric(S: Ball) return Boolean is
  begin
    return False;
  end IsNumeric;

  function IsZero(S: Ball) return Boolean is
  begin
    return (IsZero(S.C) and then IsZero(S.R));
  end IsZero;

  procedure SetZero(S: in out Ball) is
  begin
    S := BallZero;
  end SetZero;

  function Copy(S: Ball) return Ball is
    T: constant Ball := S;
  begin
    return T;
  end Copy;

  function Approx(S: Ball) return Flt is
  begin
    return Approx(S.C);
  end Approx;

  function FPut(S: Ball) return FltVector is
  begin
    return ApproxPut(S.C);
  end FPut;

  function FGet(V: FltVector) return Ball is
    O: constant Rep := RepZero;
  begin
    return (ApproxGet(V),O);
  end FGet;

  function Scal(K: Integer) return Ball is
    O: constant Rep := RepZero;
  begin
    return (Val(K),O);
  end Scal;

  function Scal(R: Flt) return Ball is
    O: constant Rep := RepZero;
  begin
    return (Val(R),O);
  end Scal;

  function Scal(R1,R2: Rep) return Ball is
    Sum: constant Rep := R1+R2;
    Eps: constant Rep := Sum+(-R1-R2);
  begin
    if R1 >= R2 then
      return (Half*Sum,Half*((R1-R2)+Eps));
    end if;
    return (Half*Sum,Half*((R2-R1)+Eps));
  end Scal;

  function Scal(R1,R2: Flt) return Ball is
  begin
    return Scal(Val(R1),Val(R2));
  end Scal;

  procedure Swap(S1,S2: in out Ball) is
  begin
    Swap(S1.C,S2.C);
    Swap(S1.R,S2.R);
  end Swap;

---------------------

  function Sup(S: Ball) return Flt is
  begin
    return Upper(S.C+S.R);
  end Sup;

  function Inf(S: Ball) return Flt is
  begin
    return Lower(-(S.R-S.C));
  end Inf;

  procedure Enclosure(S: in out Ball; Binary_Digs: in Positive) is
  begin
    if Binary_Digs<Precision then
      Set_Precision(RepZero,Binary_Digs);
      declare
        C: constant Rep := S.C;
        R: constant Rep := S.R+(C-S.C);
      begin
        Set_Precision(RepZero,Precision);
        S := (C,R);
      end;
    end if;
  end Enclosure;

  function RSup(S: Ball) return Rep is
  begin
    return S.R+S.C;
  end RSup;

  function RInf(S: Ball) return Rep is
  begin
    return -(S.R-S.C);
  end RInf;

  pragma Inline (RSup,RInf);

  function Disjoint(S1,S2: Ball) return Boolean is
  begin
    if RSup(S1) < RInf(S2) then return True; end if;
    if RInf(S1) > RSup(S2) then return True; end if;
    raise Undefined;
  end Disjoint;

  function "<"(S1,S2: Ball) return Boolean is
  begin
    if RSup(S1) < RInf(S2) then return True; end if;
    if RInf(S1) < RSup(S2) then raise Undefined; end if;
    return False;
  end "<";

  function "<="(S1,S2: Ball) return Boolean is
  begin
    if RSup(S1) <= RInf(S2) then return True; end if;
    if RInf(S1) <= RSup(S2) then raise Undefined; end if;
    return False;
  end "<=";

  function ">"(S1,S2: Ball) return Boolean is
  begin
    if RInf(S1) > RSup(S2) then return True; end if;
    if RSup(S1) > RInf(S2) then raise Undefined; end if;
    return False;
  end ">";

  function ">="(S1,S2: Ball) return Boolean is
  begin
    if RInf(S1) >= RSup(S2) then return True; end if;
    if RSup(S1) >= RInf(S2) then raise Undefined; end if;
    return False;
  end ">=";

  function Min(S1,S2: Ball) return Ball is
    U1: constant Rep := RSup(S1);
    L2: constant Rep := RInf(S2);
    S: Ball;
  begin
    if U1 <= L2 then S := S1; return S; end if;
    declare
      L1: constant Rep := RInf(S1);
      U2: constant Rep := RSup(S2);
    begin
      if U2 <= L1 then S := S2; return S; end if;
      return Scal(Min(L1,L2),Min(U1,U2));
    end;
  end Min;

  function Max(S1,S2: Ball) return Ball is
    U1: constant Rep := RSup(S1);
    L2: constant Rep := RInf(S2);
    S: Ball;
  begin
    if U1 <= L2 then S := S2; return S; end if;
    declare
      L1: constant Rep := RInf(S1);
      U2: constant Rep := RSup(S2);
    begin
      if U2 <= L1 then S := S1; return S; end if;
      return Scal(Max(L1,L2),Max(U1,U2));
    end;
  end Max;

----------------------------

  function Ball0(S: Ball) return Ball is
    O: constant Rep := RepZero;
  begin
    return (O,abs(S.C)+S.R);
  end Ball0;

  function Ball0(R: Flt) return Ball is
    O: constant Rep := RepZero;
  begin
    return (O,Val(abs(R)));
  end Ball0;

  function Center(S: Ball) return Ball is
    T: constant Ball := (S.C,RepZero);
  begin
    return T;
  end Center;

  procedure ResetCenter(S: in out Ball) is
  begin
    S.C := RepZero;
  end ResetCenter;

  function Subset(S1,S2: Ball) return Boolean is
    R: constant Rep := abs(S1.C-S2.C);
  begin
    if (S1.R+R) <= S2.R then return True; end if;
    if (S2.R-R) >= S1.R then raise Undefined; end if;
    return False;
  end Subset;

  function Union(S1,S2: Ball) return Ball is
  begin
    return Scal(Min(RInf(S1),RInf(S2)),Max(RSup(S1),RSup(S2)));
  end Union;

  function Intersection(S1,S2: Ball) return Ball is
    R1: constant Rep := Max(RInf(S1),RInf(S2));
    R2: constant Rep := Min(RSup(S1),RSup(S2));
  begin
    if R2<R1 then raise Undefined; end if;
    return Scal(R1,R2);
  end Intersection;

--------------------------

  function Norm(S: Ball) return Ball is
    T: Ball := S;
  begin
    if T.C < T.R then
      T.C := -T.C;
      if T.C < T.R then
        T.C := Half*(abs(T.C)+T.R);
        T.R := T.C;
      end if;
    end if;
    return T;
  end Norm;

  function MaxNorm(S: Ball) return Flt is
  begin
    return Upper(abs(S.C)+S.R);
  end MaxNorm;

  procedure Neg(S: in out Ball) is
  begin
    S.C := -S.C;
  end Neg;

  function "-"(S: Ball) return Ball is
    T: constant Ball := (-S.C,S.R);
  begin
    return T;
  end "-";

  procedure Add(S: in Ball; T: in out Ball) is
    Sum: constant Rep := S.C+T.C;
    Eps: constant Rep := Sum+(-S.C-T.C);
  begin
    T.C := Sum;
    T.R := (S.R+T.R)+Eps;
  end Add;

  function "+"(S,T: Ball) return Ball is
    Sum: constant Rep := S.C+T.C;
    Eps: constant Rep := Sum+(-S.C-T.C);
  begin
    return (Sum,S.R+T.R+Eps);
  end "+";

  procedure Sub(S: in Ball; T: in out Ball) is
    Dif: constant Rep := T.C-S.C;
    Eps: constant Rep := Dif+(S.C-T.C);
  begin
    T.C := Dif;
    T.R := (S.R+T.R)+Eps;
  end Sub;

  function "-"(S,T: Ball) return Ball is
    Dif: constant Rep := S.C-T.C;
    Eps: constant Rep := Dif+(T.C-S.C);
  begin
    return (Dif,S.R+T.R+Eps);
  end "-";

  procedure Mult(R: in Flt; S: in out Ball) is
    Pro: constant Rep := R*S.C;
    Eps: constant Rep := Pro+(-R)*S.C;
  begin
    S.C := Pro;
    S.R := abs(R)*S.R+Eps;
  end Mult;

  function "*"(R: Flt; S: Ball) return Ball is
    Pro: constant Rep := R*S.C;
    Eps: constant Rep := Pro+(-R)*S.C;
  begin
    return (Pro,abs(R)*S.R+Eps);
  end "*";

  function Sqr(S: Ball) return Ball is
    Pro: constant Rep := S.C*S.C;
    Eps: constant Rep := Pro+S.C*(-S.C);
  begin
    return (Pro,(Two*abs(S.C)+S.R)*S.R+Eps);
  end Sqr;

  procedure Mult(S: in Ball; T: in out Ball) is
    Pro: constant Rep := S.C*T.C;
    Eps: constant Rep := Pro+S.C*(-T.C);
  begin
    T.R := abs(S.C)*T.R+S.R*(abs(T.C)+T.R)+Eps;
    T.C := Pro;
  end Mult;

  function "*"(S,T: Ball) return Ball is
    Pro: constant Rep := S.C*T.C;
    Eps: constant Rep := Pro+S.C*(-T.C);
  begin
    return (Pro,abs(S.C)*T.R+S.R*(abs(T.C)+T.R)+Eps);
  end "*";

  procedure AddProd(S1,S2: in Ball; S3: in out Ball) is
  begin
    Add(S1*S2,S3);
  end AddProd;

  function Inv(S: Ball) return Ball is
    R:   constant Rep := S.R-abs(S.C);
    Quo: constant Rep := RepOne/S.C;
    Eps: constant Rep := Quo+RepNegOne/S.C;
  begin
    if Sign(R) >= 0 then raise Constraint_Error; end if;  -- division by zero
    return (Quo,(S.R/abs(S.C))/(-R)+Eps);
  end Inv;

  function "/"(S,T: Ball) return Ball is
  begin
    return S*Inv(T);
  end "/";

  function "**"(S: Ball; I: Integer) return Ball is
  begin
    if I>0 then
      declare
        J: Integer := I;
        X: Ball := S;
        Y: Ball;
      begin
        if (J rem 2)=0 then
          Y := BallOne;
        else
          Y := X;
        end if;
        J := J/2;
        while J>0 loop
          X := Sqr(X);
          if (J rem 2)>0 then
            Mult(X,Y);
          end if;
          J := J/2;
        end loop;
        return Y;
      end;
    elsif I<0 then
      return Inv(S)**(-I);
    else
      return (Val(1),Val(0));
    end if;
  end "**";

--------------

  function Sqrt(S: Ball) return Ball is
--- assumig IEEE
  begin
    if IsZero(S) then return S; end if;
    declare
      Ru: constant Rep := Sqrt(S.C+S.R);
      Rd: constant Rep := -((S.R-S.C)/Sqrt(S.C-S.R));
    begin
      return Scal(Rd,Ru);
    end;
  end Sqrt;

  function ExpUp(R: Rep) return Rep is
    Iter: constant Integer := 32;
    K: Integer := Sign(R);
    X: Rep := R;
    Y: Rep := RepOne;
  begin
    if K>0 then
      Y := Val(257)/Val(256);     --  assumes Iter >= 32
    else
      Y := RepOne;
      if K=0 then return Y; end if;
    end if;
    K := 0;
    while abs(X)>RepEighth loop
      K := K+1;
      X := Half*X;
    end loop;
    for I in reverse 1 .. Iter loop
      Y := RepOne+Y*(X/Val(I));
    end loop;
    for I in 1 .. K loop
      Y := Y*Y;
    end loop;
    return Y;
  end ExpUp;

  function ExpDown(R: Rep) return Rep is
  begin
    return -(RepNegOne/ExpUp(-R));
  end ExpDown;

  function Exp(S: Ball) return Ball is
  begin
    return Scal(ExpDown(RInf(S)),ExpUp(RSup(S)));
  end Exp;

  procedure CosSin(S: in Ball; Sc,Ss: in out Ball) is
    Iter: constant Integer := 8;
    Eps: constant Rep := RepOne/Val(1024);
  begin
    if IsZero(S) then
      Sc := BallOne;
      Ss := S;
    else
      declare
        K: Integer := 0;
        R: Flt := MaxNorm(S);
        X,T: Ball;
      begin
        while R>Half loop
          K := K+1;
          R := Half*R;
        end loop;
        X := (Half**K)*S;
        T := Sqr(X);
        Sc := (RepOne-Eps,Eps);
        Ss := Sc;
        for I in reverse 1 .. Iter loop
          Sc := BallOne-Sc*T/Scal((2*I)*(2*I-1));
          Ss := BallOne-Ss*T/Scal((2*I)*(2*I+1));
        end loop;
        Mult(X,Ss);
        for I in 1 .. K loop
          T := Sc;
          Sc := BallOne-Two*Sqr(Ss);
          Mult(Two*T,Ss);
        end loop;
      end;
    end if;
  end CosSin;

  function Cos(S: Ball) return Ball is
    Sc,Ss: Ball;
  begin
    CosSin(S,Sc,Ss);
    return Sc;
  end Cos;

  function Sin(S: Ball) return Ball is
    Sc,Ss: Ball;
  begin
    CosSin(S,Sc,Ss);
    return Ss;
  end Sin;

  function Log(S: Ball) return Ball is
    LogS0: Ball;
  begin
    if S=BallOne then LogS0 := BallZero; return LogS0; end if;
    declare
      Iter: constant Integer := 3;
      S0,X,Y: Ball;
    begin
      LogS0 := (Num_Log(S.C),RepZero); -- guess
      S0 := Exp(LogS0);
      X := (S0-S)/S0;                   -- Log(S) = Log(S0)+Log(1-X)
      Y := BallOne-Scal(MaxNorm(X));
      if Sign(RInf(Y)) <= 0 then raise Constraint_Error; end if;
      Y := Ball0(Inv(Flt(Iter+1)*Y));
      for I in reverse 1 .. Iter loop
        Y := Inv(Scal(I))+X*Y;
      end loop;
      return LogS0-X*Y;
    end;
  end Log;

  function Pi return Ball is
--- Bailey, 1997
    Iter: constant Integer := 20;
    S: constant Ball := Inv(Scal(16));
    P: Ball := (RepZero,RepOne);
  begin
    for K in reverse 0 .. Iter loop
      P := S*P+(Inv(Scal(8*K+1))-Inv(Scal(16*K+8))-Inv(Scal(32*K+20))-Inv(Scal(32*K+24)));
    end loop;
    return Scal(4)*P;
  end Pi;

--------------------

  procedure ShowType(S: in Ball) is
  begin
    Show0("Scalar type is Ball");
  end ShowType;

  procedure Show1(N: in String; S: in Ball; NewLine: in Boolean := True) is
    Digs: constant Natural := 0; -- use default
  begin
    Put(N & DecStr(S.C,Digs));
    if not IsZero(S.R) then
      Put("  [" & DecStr(S.R,Digs) & "]");
    end if;
    if NewLine then New_Line; end if;
  end Show1;

  procedure Show2(N: in String; S1,S2: in Ball; NewLine: in Boolean := True) is
  begin
    Show1(N,S1,False);
    Show1("   ",S2,NewLine);
  end Show2;

  procedure Put(F: in File_Type; S: in Ball; Decimal: in Boolean := False) is
  begin
    if Decimal then
      Put_Line(F,DecStr(S.C));
    elsif IsZero(S.R) then
      Put_Line(F,HexStr(S.C));
    else
      Put_Line(F,HexStr(S.C) & "," & HexStr(S.R));
    end if;
  end Put;

  function Enlarge(R: in Rep) return Ball is
    S: Ball := (R,RepZero);
  begin
    if Rounded_Last_Conversion then
      S.R := (R+RepTiny)-R;        --- cheap
    end if;
    return S;
  end Enlarge;

  procedure Get(F: in File_Type; S: in out Ball; Decimal: in Boolean := False) is
    use Ada.Strings.Unbounded;
    U: Unbounded_String := Null_Unbounded_String;
    E: Ball;
  begin
    while Length(U)=0 loop
      Text_IO.Get_Line(F,U);
    end loop;
    declare
      V: constant String := To_String(U);
      N: Integer := V'First;
    begin
      while N <= V'Last loop
        exit when V(N)=',';
        N := N+1;
      end loop;
      if Decimal then
        S := (ValDec(V(V'First .. N-1)),RepZero);
      else
        S := Enlarge(ValHex(V(V'First .. N-1)));
        if N<V'Last then
          E := Enlarge(ValHex(V(N+1 .. V'Last)));
          Add(Ball0(E),S);
        end if;
      end if;
    end;
  end Get;

  procedure Write(Name: in String; S: in Ball; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Writing " & Name);
    Create(F,Out_File,Name);
    Put(F,S,Decimal);
    Close(F);
  end Write;

  procedure Read(Name: in String; S: in out Ball; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Reading " & Name);
    Open(F,In_File,Name);
    Get(F,S,Decimal);
    Close(F);
  end Read;

  procedure Set_Active_Mode(Dummy: in Ball) is
  begin
    Show("balls active",2);
    Active := True;
    Set_Precision(RepZero,Precision);
    Set_Rounding_Up(RepZero);
  end Set_Active_Mode;

  procedure Set_Wait_Mode(Dummy: in Ball) is
  begin
    Show("balls waiting",2);
    Active := False;
  end Set_Wait_Mode;

  procedure Set_New_Precision(Dummy: in Ball; Binary_Digs: in Positive) is
  begin
    if Binary_Digs<Precision then
      Put_Line("sorry, will not decrease precision");
    else
      Precision := Binary_Digs;
    end if;
    if Active then
      Set_Precision(RepZero,Precision);
    end if;
  end Set_New_Precision;

  function Get_Precision(S: Ball) return Positive is
  begin
    return Get_Precision(S.C);
  end Get_Precision;

begin

  Set_Active_Mode(BallZero);

end Balls;
