with Ints.IO, Messages;
use Ints.IO, Messages;

package body Contraction is

  ScalZero: constant Scalar := Scal(0);
  ScalOne:  constant Scalar := Scal(1);
  Trunc:   constant Boolean := IsNumeric(ScalZero);

  function NewDeg(K: Positive) return Boolean is
    D: constant Integer := Deg(K);
  begin
    if D=0 then
      Error("Contraction.NewDeg error");
    end if;
    return D>Deg(K-1);
  end NewDeg;

  function HighNorm(Template: Fun; K0: Positive; Target: Radius) return Radius is
    K: Integer := K0;
    R,Rk: Flt := Zero;
    F,G: Fun;
  begin
    if not Trunc then
      loop
        SetZero(Template,F);
        AddHigher(K,ScalOne,F);
        Op(F,G);
        Rk := Sup(Norm(G));
        Show("HighNorm:" & Strng(K,4) & "+",Rk,1);
        R := RMax(R,Rk);
        K := K+1;
        exit when (R>Target) or else NewDeg(K);
      end loop;
    end if;
    return R;
  end HighNorm;

  function OperatorNorm(Template: Fun; Small,Target: Radius) return Scalar is
    K,D: Integer := 1;
    R,Rk: Flt := Zero;
    S: Scalar;
    F,G: Fun;
    function HiNorm is new HighNorm(Op => HighOp);
  begin
    TraceEnter("Contraction.OperatorNorm");
    loop
      SetZero(Template,F);
      AddMode(K,ScalOne,F);
      S := Norm(F);
      LowOp(F,G);
      Rk := Sup(Norm(G)/S);
      Show("OpNorm:  " & Strng(K,4) & " ",Rk,1);
      R := RMax(R,Rk);
      K := K+1;
      if Rk<Small and then HighDeg(K) then
        Rk := HiNorm(Template,K,Target);
        exit when Rk<Target;
      end if;
    end loop;
    R := RMax(R,Rk);
    Show("OpNorm:   ****",R,1);
    TraceLeave;
    return Ball0(R);
  end OperatorNorm;

  procedure NumOpMatrix(Template: Fun; A: in out Matrix) is
--- use only with numeric scalars
    U: Vector(A'Range(2));
    V: Vector(A'Range(1));
    F,G: Fun;
  begin
    TraceEnter("Contraction.NumOpMatrix");
    for K in U'Range loop
      Show("NumOpMatrix: column" & Strng(K,3),3);
      SetZero(U);
      U(K) := ScalOne;
      UnPack(U,Template,F);
      Op(F,G);
      Pack(G,V);
      SetColumn(K,V,A);
    end loop;
    TraceLeave;
  end NumOpMatrix;

  procedure DiagProd(W: in Vector; F: in Fun; G: in out Fun) is
    V: Vector(W'Range);
  begin
    Pack(F,V);
    if IsZero(V) then
      SetZero(F,G);
    else
      for K in V'Range loop Mult(W(K),V(K)); end loop;
      UnPack(V,F,G);
    end if;
  end DiagProd;

  procedure Prod(A: in Matrix; F: in Fun; G: in out Fun) is
    V: Vector(A'Range(2));
  begin
    Pack(F,V);
    if IsZero(V) then
      SetZero(F,G);
    else
      Mult(A,V);
      UnPack(V,F,G);
    end if;
  end Prod;

  function OpNorm(Template: Fun; M: Matrix) return Scalar is
    N,S: Scalar;
    V: Vector(M'Range(2));
    G: Fun;
    function Norm(V: Vector) return Scalar is
    begin
      UnPack(V,Template,G);
      return Norm(G);
    end Norm;
  begin
    TraceEnter("Contraction.OpNorm");
    N := ScalZero;
    for K in V'Range loop
      SetZero(V);
      V(K) := ScalOne;
      S := Norm(V);
      Mult(M,V);
      S := Norm(V)/S;
      N := Max(S,N);
    end loop;
    Show("||M|| =",Sup(N),1);
    TraceLeave;
    return N;
  end OpNorm;

  procedure NumDerMatrix(DY: in DMap; P: in Param; F: in Fun; A: in out Matrix) is
    procedure DerY(G: in Fun; H: in out Fun) is
    begin
      H := G;
      DY(P,F,H);
    end DerY;
    procedure NumDerY is new NumOpMatrix(Op => DerY);
  begin
    NumDerY(F,A);
  end NumDerMatrix;

  procedure NumContrMatrix(DY: in DMap; P: in Param; F: in Fun; A: in out Matrix) is
--- use only with numeric scalars
  begin
    NumDerMatrix(DY,P,F,A);
    Add(-ScalOne,A);
    Invert(A);
    Add(ScalOne,A);
    Neg(A);
  end NumContrMatrix;

  procedure Contract(Y: in Map; P: in Param; F: in Fun; A: in Matrix; H: in out Fun; Err: out Flt) is
    AH,F1,F2: Fun;
  begin
    Prod(A,H,AH);
    Sum(H,AH,F1);
    Add(F,F1);
    Y(P,F1,F2);
    Sub(F2,F1);
    Err := Sup(Norm(F1));
    Sub(F,F2);
    Diff(F2,AH,H);
  end Contract;

  procedure Contract0(Y: in Map; P: in Param; A: in Matrix; F: in out Fun; Err: out Flt) is
    F0: Fun := F;
    H,AH: Fun;
  begin
    Y(P,F0,F);
    Diff(F,F0,H);
    Err := Sup(Norm(H));
    Prod(A,H,AH);
    Add(AH,F);
  end Contract0;

  procedure NumFix(Y: in Map; DY: in DMap; P: in Param; F: in out Fun; A: in out Matrix; Eps: in Radius) is
    SaveMinutes: constant Integer := 30;
    Iter: constant Integer := 500;
    MaxBad: constant Integer := 10;
    MaxErr: constant Flt := 1.0E+9;
    Bad: Integer := 0;
    MinErr,LastErr: Flt := MaxErr;
    Err: Flt;
    T: Timer := Reset_Timer;
    G: Fun := F;
  begin
    TraceEnter("Contraction.NumFix");
    for I in 1 .. Iter loop
      NumContrMatrix(DY,P,G,A);
      Contract0(Y,P,A,G,Err);
      Show("NumFix: " & Strng(I,4),Err,1);
      if Err<LastErr then
        Bad := 0;
        if Err<MinErr then
          F := G;
          if Err<Eps then TraceLeave; return; end if;
          if Minutes(T)>SaveMinutes then
            Write("numfix-out.data",F);
            T := Reset_Timer;
          end if;
          MinErr := Err;
        end if;
      else
        Bad := Bad+1;
        exit when (Err>MaxErr) or (Bad>MaxBad);
      end if;
      LastErr := Err;
    end loop;
    Show("Contraction.NumFix getting tired");
    TraceLeave;
  end NumFix;

  function DContractNorm(DY: DMap; P: Param; F: Fun; A: Matrix; Target: Radius) return Scalar is
    N: constant Positive := A'Last(2);
    Small: constant Radius := Quarter;
    AG: Fun;
    function HighDeg(K: Positive) return Boolean is
    begin
      return (K>N) and then NewDeg(K);
    end HighDeg;
    procedure DCLow(G: in Fun; H: in out Fun) is
    begin
      Prod(A,G,AG);
      Sum(G,AG,H);
      DY(P,F,H);
      Sub(AG,H);
    end DCLow;
    procedure DCHigh(G: in Fun; H: in out Fun) is
    begin
      H := G;
      DY(P,F,H);
    end DCHigh;
    function DCNorm is new OperatorNorm(HighDeg => HighDeg, LowOp => DCLow, HighOp => DCHigh);
  begin
    return DCNorm(F,Small,Target);
  end DContractNorm;

  procedure CheckFix(Y: in Map; DY: in DMap; P: in Param; A: in Matrix; F: in out Fun; KMax: in Radius) is
    N: constant Scalar := ScalOne+OpNorm(F,A);
    Rad,Del,K: Scalar;
    G: Fun;
  begin
    TraceEnter("Contraction.CheckFix");
    Y(P,F,G);
    Sub(F,G);
    Del := Norm(G);
    Show("||Y(G)-G|| =",Sup(Del),1);
    Rad := Del/(ScalOne-Scal(KMax));
    G := F;
    AddBall(N*Rad,G);
    K := DContractNorm(DY,P,G,A,KMax);
    if Sup(K) >= KMax then
      Error("CheckFix error: K>=KMax");
    end if;
    Rad := Del/(ScalOne-K);
    Show("adding ball of radius ",Sup(N*Rad),1);
    AddBall(N*Rad,F);
    TraceLeave;
  end CheckFix;

end Contraction;
