with Ints, Ints.IO, Messages, Flts.IO, Flts.Ops, Flts_Misc, Rounding;
use Ints, Ints.IO, Messages, Flts.IO, Flts_Misc;

package body Fouriers2 is
  procedure Add(F: in FErr; G: in out FErr) is
  begin
    for M in FreqE1 loop
      for N in FreqE2 loop
        AddUp(F(M,N),G(M,N));
      end loop;
    end loop;
  end Add;

  procedure Mult(S: in Scalar; E: in out FErr) is
    R: constant Flt := MaxNorm(S);
  begin
    for M in FreqE1 loop
      for N in FreqE2 loop
        MultUp(R,E(M,N));
      end loop;
    end loop;
  end Mult;

  procedure AddProd(S: in Scalar; F: in FErr; G: in out FErr) is
    R: constant Flt := MaxNorm(S);
  begin
    for M in FreqE1 loop
      for N in FreqE2 loop
        AddProdUp(R,F(M,N),G(M,N));
      end loop;
    end loop;
  end AddProd;

  function Norm(E: FErr) return Scalar is
    R: Flt := Zero;
  begin
    for M in FreqE1 loop
      for N in FreqE2 loop
        AddUp(E(M,N),R);
      end loop;
    end loop;
    return Ball0(R);
  end Norm;

------------------------------

  function IsZero(F: Fourier2) return Boolean is
  begin
    return IsZero(F.C) and then (Trunc or else IsZero(F.E));
  end IsZero;

  function IsPoly(F: Fourier2) return Boolean is
  begin
    return Trunc or else IsZero(F.E);
  end IsPoly;

  procedure FillZero(F: in out Fourier2) is
  begin
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
  end FillZero;

  procedure SetZero(T: in FType; F: in out Fourier2) is
  begin
    F.T := T;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
  end SetZero;

  procedure SetZero(Template: in Fourier2; F: in out Fourier2) is
  begin
    F.T := Template.T;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
  end SetZero;

  procedure SetZero(P: in ParPair; R: in RadPair; F: in out Fourier2) is
  begin
    F.T.P := P;
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
  end SetZero;

  procedure Powers(S: in Scalar; W: in out ScalVec) is
    M: Integer := 1;
  begin
    W(0) := ScalOne;
    W(1) := S;
    for K in 2 .. W'Last loop
      W(K) := W(K-M)*W(M);
      if 2*M=K then M := K; end if;
    end loop;
    for K in W'First .. -1 loop
      W(K) := Inv(W(-K));
    end loop;
  end Powers;

  function NotExp(X: Flt) return Scalar is
  begin
    return ScalOne+Scal(X);
  end NotExp;

  procedure UpdateNormFacs(R: in RadPair) is
--- uses global LastPair, NormFacs1, NormFacs2, NormFacs
  begin
    if R(1)=LastPair(1) then
      if R(2)=LastPair(2) then return; end if;
    else
      Powers(NotExp(R(1)),NormFacs1);
    end if;
    if R(2) /= LastPair(2) then
      if R(2)=R(1) then
        NormFacs2 := NormFacs1;
      else
        Powers(NotExp(R(2)),NormFacs2);
      end if;
    end if;
    for M in FreqE1 loop
      for N in FreqE2 loop
        NormFacs(M,N) := NormFacs1(M)*NormFacs2(N);
      end loop;
    end loop;
    LastPair := R;
  end UpdateNormFacs;

  function Norm(R: RadPair; C: FCoeff) return Scalar is
    S: Scalar := ScalZero;
  begin
    UpdateNormFacs(R);
    for M in Freq1 loop
      for N in Freq2 loop
        AddProd(NormFacs(M,N),Norm(C(M,N)),S);
      end loop;
    end loop;
    return S;
  end Norm;

  function Norm(R: RadPair; F: Fourier2) return Scalar is
  begin
    if Trunc then
      return Norm(R,F.C);
    else
      return Norm(R,F.C)+Norm(F.E);
    end if;
  end Norm;

  function Norm(F: Fourier2) return Scalar is
  begin
    return Norm(F.T.R,F);
  end Norm;

  function IProd(C1,C2: FCoeff) return Scalar is
    S: Scalar := ScalZero;
  begin
    for M in reverse Freq1 loop
      for N in reverse Freq2 loop
        AddProd(C1(M,N),C2(M,N),S);
      end loop;
    end loop;
    AddProd(C1(0,0),C2(0,0),S);
    return S;
  end IProd;

  function IProd(F,G: in Fourier2) return Scalar is
--- can be improved
  begin
    if IsPoly(G) then
      if IsPoly(F) then return IProd(F.C,G.C); end  if;
      declare
        D: constant Integer := IMax(EffDeg1(G.C),EffDeg2(G.C));
        H: Fourier2 := F;
      begin
        DistErr(D,H);
        return IProd(H.C,G.C);
      end;
    elsif not IsPoly(F) then
      Error("Fouriers2.IProd error: not implemented");
    end if;
    return IProd(G,F);
  end IProd;

  function Norm2(F: Fourier2) return Scalar is
  begin
    return Sqrt(IProd(F,F));
  end Norm2;

  function CosSin(P: Parity; S: Scalar) return Scalar is
  begin
    if P=0 then return Cos(S); else return Sin(S); end if;
  end CosSin;

  function Val(X,Y: Scalar; F: Fourier2) return Scalar is
    P1: constant Parity := F.T.P(1);
    P2: constant Parity := F.T.P(2);
    V: Scalar := ScalZero;
    S: Scalar;
    C: FCoeff renames F.C;
  begin
    for M in Freq1 loop
      S := CosSin(P1,Flt(M)*X);
      for N in Freq2 loop
        V := V+C(M,N)*S*CosSin(P2,Flt(N)*Y);
      end  loop;
    end loop;
    if not Trunc then V := V+Norm(F.E); end if;
    return V;
  end  Val;

  function Coeff(M,N: Natural; R: RadPair; E: FErr) return Scalar is
--- too slow for use in loops
    M0: constant Natural := IMin(M,DegE1);
    N0: constant Natural := IMin(N,DegE2);
    S: Flt := Zero;
  begin
    for M1 in 0 .. M0 loop
      for N1 in 0 .. N0 loop
        AddUp(E(M1,N1),S);
      end loop;
    end loop;
    if S=Zero then
      return Ball0(S);
    else
      UpdateNormFacs(R);
      return Inv(NormFacs(M,N))*Ball0(S);
    end if;
  end Coeff;

  function Coeff(M,N: Natural; F: Fourier2) return Scalar is
--- too slow for use in loops
    C,E: Scalar := ScalZero;
  begin
    if (M <= Deg1) and then (N <= Deg2) then C := F.C(M,N); end if;
    if not Trunc then E := Coeff(M,N,F.T.R,F.E); end if;
    return C+E;
  end Coeff;

  procedure AddBall(M,N: in Natural; R: in Flt; E: in out FErr) is
  begin
    AddUp(abs(R),E(IMin(M,DegE1),IMin(N,DegE2)));
  end AddBall;

  procedure AddCoeff(M,N: in Natural; S: in Scalar; F: in out Fourier2) is
  begin
    if (M <= Deg1) and then (N <= Deg2) then
      Add(S,F.C(M,N));
    elsif not (Trunc or else IsZero(S)) then
      UpdateNormFacs(F.T.R);
      AddBall(M,N,NormFacs(M,N)*S,F);
    end if;
  end AddCoeff;

  procedure AddBall(M,N: in Natural; S: in Scalar; F: in out Fourier2) is
  begin
    if not Trunc then
      AddBall(M,N,MaxNorm(S),F.E);
    end if;
  end AddBall;

  procedure AddBall(S: in Scalar; F: in out Fourier2) is
  begin
    if not Trunc then
      AddUp(MaxNorm(S),F.E(0,0));
    end if;
  end AddBall;

----------------------------

  procedure Neg(F: in out Fourier2) is
  begin
    Neg(F.C);
  end Neg;

  function "-"(F: Fourier2) return Fourier2 is
    G: Fourier2 := F;
  begin
    Neg(G.C);
    return G;
  end "-";

  procedure CheckType(T1: in FType; T2: in out FType; SameParity: out Boolean) is
  begin
    SameParity := True;
    if T1 /= T2 then
      SameParity := (T1.P=T2.P);
      T2.R := Min(T1.R,T2.R);
    end if;
  end CheckType;

  procedure Add(F: in Fourier2; G: in out Fourier2) is
    OK: Boolean;
  begin
    CheckType(F.T,G.T,OK);
    if not OK then
      Error("Fouriers2.Add parity error");
    end if;
    Add(F.C,G.C);
    if not Trunc then Add(F.E,G.E); end if;
  end Add;

  procedure Sum(F,G: in Fourier2; H: in out Fourier2) is
  begin
    H := F;
    Add(G,H);
  end Sum;

  function "+"(F,G: Fourier2) return Fourier2 is
    H: Fourier2 := F;
  begin
    Add(G,H);
    return H;
  end "+";

  procedure Sub(F: in Fourier2; G: in out Fourier2) is
    OK: Boolean;
  begin
    CheckType(F.T,G.T,OK);
    if not OK then
      Error("Fouriers2.Sub parity error");
    end if;
    Sub(F.C,G.C);
    if not Trunc then Add(F.E,G.E); end if;
  end Sub;

  procedure Diff(F,G: in Fourier2; H: in out Fourier2) is
  begin
    H := F;
    Sub(G,H);
  end Diff;

  function "-"(F,G: Fourier2) return Fourier2 is
    H: Fourier2 := F;
  begin
    Sub(G,H);
    return H;
  end "-";

  procedure Mult(S: in Scalar; F: in out Fourier2) is
  begin
    if IsZero(S) then
      FillZero(F);
    else
      Mult(S,F.C);
      if not Trunc then Mult(S,F.E); end if;
    end if;
  end Mult;

  procedure Prod(S: in Scalar; F: in Fourier2; G: in out Fourier2) is
  begin
    if IsZero(S) then
      SetZero(F.T,G);
    else
      G := F;
      Mult(S,G.C);
      if not Trunc then Mult(S,G.E); end if;
    end if;
  end Prod;

  function "*"(S: Scalar; F: Fourier2) return Fourier2 is
    G: Fourier2;
  begin
    Prod(S,F,G);
    return G;
  end "*";

  procedure AddProd(S: in Scalar; F: in Fourier2; G: in out Fourier2) is
    OK: Boolean;
  begin
    if not IsZero(S) then
      CheckType(F.T,G.T,OK);
      if not OK then
        Error("Fouriers2.AddProd parity error");
      end if;
      AddProd(S,F.C,G.C);
      if not Trunc then AddProd(S,F.E,G.E); end if;
    end if;
  end AddProd;

  procedure Signs(P1,P2: in Parity; Sum,DiffPlus,DiffMinus: out Flt) is
  begin
    Sum := One;
    DiffPlus := One;
    DiffMinus := One;
    if P1=0 then
      if P2=1 then  DiffPlus := NegOne; end if;
    elsif P2=0 then
      DiffMinus := NegOne;
    else
      Sum := NegOne;
    end if;
  end Signs;

  procedure AddProdCC(F1,F2: in Fourier2; F3: in out Fourier2) is
--- assuming correct NormFacs
    MS,MD,NS,ND: Integer;
    Xs,Ys,Xd,Xdp,Xdm,Yd,Ydp,Ydm,XsYs: Flt;
    C1,C2,C12: Scalar;
    FC1: FCoeff renames F1.C;
    FC2: FCoeff renames F2.C;
    FC3: FCoeff renames F3.C;
    FE3: FErr renames F3.E;
  begin
    Signs(F1.T.P(1),F2.T.P(1),Xs,Xdp,Xdm);
    Signs(F1.T.P(2),F2.T.P(2),Ys,Ydp,Ydm);
    XsYs := Xs*Ys;
    for M1 in Freq1 loop
      for N1 in Freq2 loop
        C1 := FC1(M1,N1);
        if C1 /= ScalZero then
          for M2 in Freq1 loop
            MS := M1+M2;
            if M2<M1 then
              MD := M1-M2;
              Xd := Xdp;
            else
              MD := M2-M1;
              Xd := Xdm;
            end if;
            for N2 in Freq2 loop
              C2 := FC2(M2,N2);
              if C2 /= ScalZero then
                NS := N1+N2;
                if N2<N1 then
                  ND := N1-N2;
                  Yd := Ydp;
                else
                  ND := N2-N1;
                  Yd := Ydm;
                end if;
                C12 := Quarter*(C1*C2);       --- factor 1/4
                Add((Xd*Yd)*C12,FC3(MD,ND));
                if MS <= Deg1 then
                  Add((Xs*Yd)*C12,FC3(MS,ND));
                  if NS <= Deg2 then
                    Add(XsYs*C12,FC3(MS,NS));
                  elsif not Trunc then
                    AddUp(MaxNorm(NormFacs(MS,NS)*C12),FE3(MS,NS));
                  end if;
                elsif not Trunc then
                  AddUp(MaxNorm(NormFacs(MS,ND)*C12),FE3(MS,ND));
                  AddUp(MaxNorm(NormFacs(MS,NS)*C12),FE3(MS,NS));
                end if;
                if NS <= Deg2 then
                  Add((Xd*Ys)*C12,FC3(MD,NS));
                elsif not Trunc then
                  AddUp(MaxNorm(NormFacs(MD,NS)*C12),FE3(MD,NS));
                end if;
              end if;
            end loop;
          end loop;
        end if;
      end loop;
    end loop;
  end AddProdCC;

  procedure AddProdCE(F1,F2: in Fourier2; F3: in out Fourier2) is
--- assuming correct NormFacs
    MS,MD,NS,ND: Integer;
    E1,E2: Flt;
    C1,WMD,WND: Scalar;
    FC1: FCoeff renames F1.C;
    FE2: FErr renames F2.E;
    FE3: FErr renames F3.E;
  begin
    if IsZero(FE2) then return; end if;
    for M1 in Freq1 loop
      for N1 in Freq2 loop
        C1 := FC1(M1,N1);
        if C1 /= ScalZero then
          E1 := MaxNorm(Quarter*NormFacs(M1,N1)*C1);  --- factor 1/4
          for M2 in FreqE1 loop
            MS := IMin(M2+M1,DegE1);
            MD := IMax(M2-M1,0);
            WMD := Sqr(NormFacs1(-IMin(M1,M2)));
            for N2 in FreqE2 loop
              E2 := FE2(M2,N2);
              if E2 /= Zero then
                NS := IMin(N2+N1,DegE2);
                ND := IMax(N2-N1,0);
                WND := Sqr(NormFacs2(-IMin(N1,N2)));
                AddProdUp(E1,E2,FE3(MS,NS));
                AddProdUp(Sup(E1*WND),E2,FE3(MS,ND));
                AddProdUp(Sup(E1*WMD),E2,FE3(MD,NS));
                AddProdUp(Sup(E1*(WND*WMD)),E2,FE3(MD,ND));
              end if;
            end loop;
          end loop;
        end if;
      end loop;
    end loop;
  end AddProdCE;

  procedure AddProdEE(F1,F2: in Fourier2; F3: in out Fourier2) is
--- assuming correct NormFacs
    MS,NS: Integer;
    E1,E2: Flt;
    WMD,WND: Scalar;
    FE1: FErr renames F1.E;
    FE2: FErr renames F2.E;
    FE3: FErr renames F3.E;
  begin
    if IsZero(FE2) then return; end if;
    for M1 in FreqE1 loop
      for N1 in FreqE2 loop
        E1 := FE1(M1,N1);
        if E1 /= Zero then
          MultUp(Quarter,E1);    --- factor 1/4
          for M2 in FreqE1 loop
            MS := IMin(M2+M1,DegE1);
            WMD := Sqr(NormFacs1(-IMin(M1,M2)));
            for N2 in FreqE2 loop
              E2 := FE2(M2,N2);
              if E2 /= Zero then
                NS := IMin(N2+N1,DegE2);
                WND := Sqr(NormFacs2(-IMin(N1,N2)));
                AddProdUp(E1,E2,FE3(MS,NS));
                AddProdUp(Sup(E1*WND),E2,FE3(MS,0));
                AddProdUp(Sup(E1*WMD),E2,FE3(0,NS));
                AddProdUp(Sup(E1*(WND*WMD)),E2,FE3(0,0));
              end if;
            end loop;
          end loop;
        end if;
      end loop;
    end loop;
  end AddProdEE;

  function "*"(P1,P2: ParPair) return ParPair is
  begin
    return ((P1(1)+P2(1)) mod 2,(P1(2)+P2(2)) mod 2);
  end "*";

  procedure Normalize(F: in out Fourier2) is
    P: ParPair renames F.T.P;
    E: FErr renames F.E;
  begin
    if P(1)=1 then
      SetZeroRow(0,F.C);
      if not Trunc then
        for N in FreqE2 loop
          AddUp(E(0,N),E(1,N));
          E(0,N) := Zero;
        end loop;
      end if;
    end if;
    if P(2)=1 then
      SetZeroColumn(0,F.C);
      if not Trunc then
        for M in FreqE1 loop
          AddUp(E(M,0),E(M,1));
          E(M,0) := Zero;
        end loop;
      end if;
    end if;
  end Normalize;

  procedure AddProd(F1,F2: in Fourier2; F3: in out Fourier2) is
    T1: FType renames F1.T;
    T2: FType renames F2.T;
    T3: FType renames F3.T;
  begin
    if T1.P*T2.P /= T3.P then
      Error("Fouriers2.AddProd parity error");
    end if;
    T3.R := Min(Min(T1.R,T2.R),T3.R);
    UpdateNormFacs(T3.R);
    AddProdCC(F1,F2,F3);
    if not Trunc then
      AddProdCE(F1,F2,F3);
      AddProdCE(F2,F1,F3);
      AddProdEE(F1,F2,F3);
     end if;
    Normalize(F3);
  end AddProd;

  procedure Mult(F: in Fourier2; G: in out Fourier2) is
    H: constant Fourier2 := G;
  begin
    SetZero(F.T.P*H.T.P,F.T.R,G);
    AddProd(F,H,G);
  end Mult;

  procedure Prod(F,G: in Fourier2; H: in out Fourier2) is
  begin
    SetZero(F.T.P*G.T.P,F.T.R,H);
    AddProd(F,G,H);
  end Prod;

  function "*"(F,G: Fourier2) return Fourier2 is
    H: Fourier2;
  begin
    SetZero(F.T.P*G.T.P,F.T.R,H);
    AddProd(F,G,H);
    return H;
  end "*";

  procedure Show1(S: in String; T: in FType) is
  begin
    if T.P(1)=0 then
      Show0(S & " cos*",False);
    else
      Show0(S & " sin*",False);
    end if;
    if T.P(2)=0 then
      Show2("cos  Rho",T.R(1),T.R(2));
    else
      Show2("sin  Rho",T.R(1),T.R(2));
    end if;
    New_Line;
  end Show1;

  procedure Show1(S: in String; F: in Fourier2) is
    W: constant Positive := 4;
  begin
    if IsZero(F) then
      Put(S & "  Zero_Fourier2");
    else
      Show1(S,F.T);
      Show1(S,F.C);
      if not Trunc then
        New_Line;
        for M in FreqE1 loop
          for N in FreqE2 loop
            if F.E(M,N) /= Zero then
              Show1(S & " Err " & Strng(M,W) & Strng(N,W),F.E(M,N));
            end if;
          end loop;
        end loop;
      end if;
    end if;
    New_Line;
  end Show1;

  procedure Show2(S: in String; F1,F2: in Fourier2) is
    W: constant Positive := 4;
  begin
    if IsZero(F1) and then IsZero(F2) then
      Put(S & "  Zero_Fouriers2");
    else
      Show1(S & "1st  ",F1.T);
      Show1(S & "2nd  ",F2.T);
      Show2(S,F1.C,F2.C);
      if not Trunc then
        New_Line;
        for M in FreqE1 loop
          for N in FreqE2 loop
            if (F1.E(M,N) /= Zero) or else (F2.E(M,N) /= Zero) then
              Show2(S & " Err " & Strng(M,W) & Strng(N,W),F1.E(M,N),F2.E(M,N));
            end if;
          end loop;
        end loop;
      end if;
    end if;
    New_Line;
  end Show2;

  function EffDeg1(C: FCoeff) return Integer is
  begin
    for M in reverse Freq1 loop
      for N in Freq2 loop
        if C(M,N) /= ScalZero then return M; end if;
      end loop;
    end loop;
    return -1;
  end EffDeg1;

  function EffDeg2(C: FCoeff) return Integer is
  begin
    for N in reverse Freq2 loop
      for M in Freq1 loop
        if C(M,N) /= ScalZero then return N; end if;
      end loop;
    end loop;
    return -1;
  end EffDeg2;

  procedure EffDegs0(C: in FCoeff; D1,D2: out Integer) is
  begin
    D1 := IMax(EffDeg1(C),0);
    D2 := IMax(EffDeg2(C),0);
  end EffDegs0;

  function EffDeg1(E: FErr) return Integer is
  begin
    for M in reverse Freq1 loop
      for N in Freq2 loop
        if E(M,N) /= Zero then return M; end if;
      end loop;
    end loop;
    return -1;
  end EffDeg1;

  function EffDeg2(E: FErr) return Integer is
  begin
    for N in reverse Freq2 loop
      for M in Freq1 loop
        if E(M,N) /= Zero then return N; end if;
      end loop;
    end loop;
    return -1;
  end EffDeg2;

  procedure EffDegs(E: in FErr; D1,D2: out Integer) is
  begin
    D1 := EffDeg1(E);
    D2 := EffDeg2(E);
  end EffDegs;

  procedure EffDegs0(E: in FErr; D1,D2: out Integer) is
  begin
    D1 := IMax(0,EffDeg1(E));
    D2 := IMax(0,EffDeg2(E));
  end EffDegs0;

  procedure Put(F: in File_Type; H: in Fourier2; Decimal: in Boolean := False) is
    D1,D2: Integer;
  begin
    Put(F,H.T.P(1));
    Put(F,H.T.P(2));
    Put(F,H.T.R(1),Decimal);
    Put(F,H.T.R(2),Decimal);
    EffDegs0(H.C,D1,D2);
    Put(F,D1);
    Put(F,D2);
    for M in 0 .. D1 loop
      for N in 0 .. D2 loop
        Put(F,H.C(M,N),Decimal);
      end loop;
    end loop;
    EffDegs0(H.E,D1,D2);
    Put(F,D1);
    Put(F,D2);
    for M in 0 .. D1 loop
      for N in 0 .. D2 loop
        Put(F,H.E(M,N),Decimal);
      end loop;
    end loop;
  end Put;

  procedure Get(F: in File_Type; H: in out Fourier2; Decimal: in Boolean := False) is
    D1,D2: Integer;
    E: Flt;
    C: Scalar;
    Warn: Boolean := False;
  begin
    FillZero(H);
    Get(F,H.T.P(1));
    Get(F,H.T.P(2));
    Get(F,H.T.R(1),Decimal);
    Get(F,H.T.R(2),Decimal);
    Get(F,D1);
    Get(F,D2);
    for M in 0 .. D1 loop
      for N in 0 .. D2 loop
        Get(F,C,Decimal);
        AddCoeff(M,N,C,H);
      end loop;
    end loop;
    Get(F,D1);
    Get(F,D2);
    for M in 0 .. D1 loop
      for N in 0 .. D2 loop
        Get(F,E,Decimal);
        AddBall(M,N,E,H.E);
      end loop;
    end loop;
  end Get;

  procedure Write(Name: in String; H: in Fourier2; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Writing " & Name);
    Create(F,Out_File,Name);
    Put(F,H,Decimal);
    Close(F);
  end Write;

  procedure Read(Name: in String; H: in out Fourier2; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Reading " & Name);
    Open(F,In_File,Name);
    Get(F,H,Decimal);
    Close(F);
  end Read;

  procedure NumRead(Name: in String; H: in out Fourier2; Decimal: in Boolean := False) is
    C: FCoeff renames H.C;
  begin
    Read(Name,H,Decimal);
    if not Trunc then
      for I in C'Range(1) loop
        for J in C'Range(2) loop
          C(I,J) := Center(C(I,J));
        end loop;
      end loop;
      H.E := Zero_Ferr;
    end if;
  end NumRead;

  procedure Plot(F: in Fourier2; L: Positive := 50) is
    use Flt_IO;
    P: constant Flt := 3.141592653589793;
    X,Y,Z: Scalar;
  begin
    for M in 0 .. L loop
      for N in 0 .. L loop
        X := P*Scal(M)/Scal(L);
        Y := P*Scal(N)/Scal(L);
        Z := Val(X,Y,F);
        Put(Approx(X),3,3,3);
        Put(Approx(Y),3,3,3);
        Put(Approx(Z),3,3,3);
        New_Line;
      end loop;
      New_Line;
    end loop;
  end Plot;

  procedure GuessRho(F: Fourier2) is
    use Flts.Ops;
    Deg: constant Integer := IMin(Deg1,Deg2);
    D0: constant Integer := Deg/4;
    D1: constant Integer := Deg-D0;
    J,N1,N2: Integer := 0;
    Q1,Q2: Flt := One;
    T: FType := F.T;
    procedure NewQuot(S0,S1: in Scalar; Q: in out Flt; N: in out Integer) is
    begin
      if (S0 /= ScalZero) and then (S1 /= ScalZero) then
        Q := Q*MaxNorm(S0)/MaxNorm(S1);
        N := N+1;
      end if;
    end NewQuot;
  begin
    for D in D0 .. D1-1 loop
      for I in 0 .. D loop
        J := D-I;
        NewQuot(F.C(I,J),F.C(I+1,J),Q1,N1);
        NewQuot(F.C(I,J),F.C(I,J+1),Q2,N2);
      end loop;
    end loop;
    T.R(1) := Log(Q1)/Flt(N1);
    T.R(2) := Log(Q2)/Flt(N2);
    Show1("guessed domain:",T);
  end GuessRho;

-----------------------

  procedure CReflect(C: in out FCoeff) is
    Swap: Boolean;
    Sig: Flt;
    S: Scalar;
  begin
    for M in Freq1 loop
      for N in Freq2 loop
        Ref(M,N,Swap,Sig);
        if not Swap then
          Mult(Sig,C(M,N));
        elsif M<N then
          S := C(M,N);
          C(M,N) := Sig*C(N,M);
          C(N,M) := Sig*S;
        end if;
      end loop;
    end loop;
  end CReflect;

  procedure EReflect(E: in out FErr) is
    Swap: Boolean;
    Sig,R: Flt;
  begin
    for M in FreqE1 loop
      for N in FreqE2 loop
        Ref(M,N,Swap,Sig);
        if Swap then
          R := E(M,N);
          E(M,N) := E(N,M);
          E(N,M) := R;
        end if;
      end loop;
    end loop;
  end EReflect;

  procedure Reflect(F: in out Fourier2) is
    Swap: Boolean;
    P: Parity;
    R: Flt;
    T: FType renames F.T;
  begin
    if Ref=Nop then return; end if;
    Ref(0,1,Swap,R);
    if Swap then
      P := T.P(1); T.P(1) := T.P(2); T.P(2) := P;
      R := T.R(1); T.R(1) := T.R(2); T.R(2) := R;
    end if;
    CReflect(F.C);
    if not Trunc then EReflect(F.E); end if;
  end Reflect;

  procedure CSymm(C: in out FCoeff) is
    Swap: Boolean;
    Sig: Flt;
  begin
    for M in Freq1 loop
      for N in Freq2 loop
        Ref(M,N,Swap,Sig);
        if not Swap then
          if Sig<Zero then C(M,N) := ScalZero; end if;
        elsif M<N then
          C(M,N) := Half*(C(M,N)+Sig*C(N,M));
          C(N,M) := Sig*C(M,N);
        end if;
      end loop;
    end loop;
  end CSymm;

  procedure ESymm(E: in out FErr) is
    Swap: Boolean;
    Sig: Flt;
  begin
    for M in FreqE1 loop
      for N in FreqE2 loop
        Ref(M,N,Swap,Sig);
        if Swap then
          E(M,N) := Half*SumUp(E(M,N),E(N,M));
          E(N,M) := E(M,N);
        end  if;
      end loop;
    end loop;
  end ESymm;

  procedure Symmetrize(F: in out Fourier2) is
    Swap: Boolean;
    Sig: Flt;
    T: FType renames F.T;
  begin
    if Ref=Nop then return; end if;
    Ref(0,1,Swap,Sig);
    if Swap then
      if T.P(1) /= T.P(2) then
        Error("Fouriers2.Symmetrize parity error");
      end if;
      T.R(1) := RMin(T.R(1),T.R(2));
      T.R(2) := T.R(1);
    end if;
    CSymm(F.C);
    if not Trunc then ESymm(F.E); end if;
  end Symmetrize;

------------------------

  function ModeCoeff(K: Positive; C: FCoeff) return Scalar is
    Swap: Boolean;
    Sig: Flt;
    M,N: Integer;
  begin
    Coord(K,M,N);
    Ref(M,N,Swap,Sig);
    if Swap then
      return C(M,N)+Sig*C(N,M);
    else
      return Copy(C(M,N));
    end if;
  end ModeCoeff;

  procedure ModeSplit(K: in Positive; C: in out FCoeff; S: in out Scalar) is
--- assumes C is symmetric
    Swap: Boolean;
    Sig: Flt;
    M,N: Integer;
  begin
    Coord(K,M,N);
    Ref(M,N,Swap,Sig);
    if Swap then
      S := C(M,N)+Sig*C(N,M);
      C(N,M) := ScalZero;
    else
      S := C(M,N);
    end if;
    C(M,N) := ScalZero;
  end ModeSplit;

  function ModeCoeff(K: Positive; F: Fourier2) return Scalar is
--- too slow use in loops, use Pack instead
    Swap: Boolean;
    Sig: Flt;
    M,N: Integer;
  begin
    Coord(K,M,N);
    Ref(M,N,Swap,Sig);
    if Swap then
      return Coeff(M,N,F)+Sig*Coeff(N,M,F);
    else
      return Coeff(M,N,F);
    end if;
  end ModeCoeff;

  procedure AddMode(K: in Positive; S: in Scalar; C: in out FCoeff) is
    Swap: Boolean;
    Sig: Flt;
    M,N: Integer;
  begin
    Coord(K,M,N);
    Ref(M,N,Swap,Sig);
    if Swap then
      Add(Half*S,C(M,N));
      Add((Sig*Half)*S,C(N,M));
    else
      Add(S,C(M,N));
    end if;
  end AddMode;

  procedure AddMode(K: in Positive; S: in Scalar; F: in out Fourier2) is
  begin
    AddMode(K,S,F.C);
  end AddMode;

  procedure AddHigher(K: in Positive; R: in Flt; E: in out FErr) is
    Swap: Boolean;
    Sig: Flt;
    M,N: Integer;
  begin
    Coord(K,M,N);
    Ref(M,N,Swap,Sig);
    if Swap then
      AddBall(M,N,Half*R,E);
      AddBall(N,M,Half*R,E);
    else
      AddBall(M,N,R,E);
    end if;
  end AddHigher;

  procedure AddHigher(K: in Positive; S: in Scalar; F: in out Fourier2) is
  begin
    AddHigher(K,MaxNorm(S),F.E);
  end AddHigher;

--------------------------------

  function SomeDeg(C: FCoeff) return Positive is
  begin
    for D in reverse 2 .. IMin(Deg1,Deg2) loop
      for M in 1 .. D-1 loop
        if not IsZero(C(M,D-M)) then return MaxInvDeg(D); end if;
      end loop;
    end loop;
    return 1;
  end SomeDeg;

  procedure Pack(C: in FCoeff; V: in out ScalVec) is
  begin
    for K in 1 .. V'Last loop
      V(K) := ModeCoeff(K,C);
    end loop;
  end Pack;

  procedure PackSplit(C: in out FCoeff; V: in out ScalVec) is
  begin
    for K in 1 .. V'Last loop
      ModeSplit(K,C,V(K));
    end loop;
  end PackSplit;

  procedure DistErr(D: in Natural; F: in out Fourier2) is
--- distribute errors up to degree D into coefficients
--- using combined degree would be better
    C: FCoeff renames F.C;
    E: FErr renames F.E;
  begin
    UpdateNormFacs(F.T.R);
    for M in 0 .. D+1 loop
      for N in 0 .. D loop
        AddUp(E(M,N),E(M,N+1));   -- sum along row M
      end loop;
    end loop;
    for N in 0 .. D+1 loop
      for M in 0 .. D loop
        declare
          Emn: Flt renames E(M,N);
          B: constant Scalar := Emn*Ball0(Inv(NormFacs(M,N)));
        begin
          Add(B,C(M,N));
          AddUp(Emn,E(M+1,N));   -- sum along column N
          Emn := Zero;
        end;
      end loop;
    end loop;
  end DistErr;

  procedure Pack(F: in Fourier2; V: in out ScalVec) is
  begin
    if IsPoly(F) then
      Pack(F.C,V);
    else
      declare
        G: Fourier2 := F;
      begin
        DistErr(Deg(V'Last),G);
        Pack(G.C,V);
      end;
    end if;
  end Pack;

  procedure PackSplit(F: in out Fourier2; V: in out ScalVec) is
  begin
    if not IsPoly(F) then
      DistErr(Deg(V'Last),F);
    end if;
    PackSplit(F.C,V);
  end PackSplit;

  procedure UnPack(V: in ScalVec; C: in out FCoeff) is
  begin
    C := Zero_FCoeff;
    for K in 1 .. V'Last loop
      AddMode(K,V(K),C);
    end loop;
  end UnPack;

  procedure UnPack(V: in ScalVec; Template: in Fourier2; F: in out Fourier2) is
  begin
    F.T := Template.T;
    UnPack(V,F.C);
    F.E := Zero_FErr;
  end UnPack;

  function UnPack(V: ScalVec; Template: Fourier2) return Fourier2 is
    F: Fourier2;
  begin
    F.T := Template.T;
    UnPack(V,F.C);
    F.E := Zero_FErr;
    return F;
  end UnPack;

--------------------------

  function AntiBumpsX(R: RadPair) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (1,1);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(2,1) :=  ScalOne;
    return F;
  end AntiBumpsX;

  function AntiBumpsY(R: RadPair) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (1,1);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(1,2) :=  ScalOne;
    return F;
  end AntiBumpsY;

  function AntiBumpsDiag(R: RadPair) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (1,1);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(1,2) :=  ScalOne;
    F.C(2,1) :=  ScalOne;
    return F;
  end AntiBumpsDiag;

  function MidPeak(R: RadPair; N: Positive) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (1,1);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(1,1) := ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end MidPeak;

  function MidHole(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2 := MidPeak(R,1);
  begin
    Neg(F);
    F.C(0,0) := F.C(0,0)+ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end MidHole;

  function EdgePeaks(R: RadPair; N: Positive) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (0,0);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(2,0) := ScalHalf;
    F.C(0,2) := -ScalHalf;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end EdgePeaks;

  function CornerPeaks(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (0,0);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(0,0) := ScalHalf;
    F.C(2,0) := ScalQuarter;
    F.C(0,2) := ScalQuarter;
    return F;
  end CornerPeaks;

  function MValley(R: RadPair; N: Positive) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (0,0);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(1,0) :=  ScalHalf;
    F.C(0,1) := -ScalHalf;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end MValley;

  function PValley(R: RadPair; N: Positive) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (0,0);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(1,0) := ScalHalf;
    F.C(0,1) := ScalHalf;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end PValley;

  function HVValley(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (0,0);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.C(0,0) := ScalQuarter;
    F.C(2,0) := ScalQuarter;
    F.C(0,2) := ScalQuarter;
    F.C(2,2) := ScalQuarter;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end HVValley;

  function MRidge(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2 := PValley(R,1);
  begin
    Neg(F);
    F.C(0,0) :=  F.C(0,0)+ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end MRidge;

  function PRidge(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2 := MValley(R,1);
  begin
    Neg(F);
    F.C(0,0) :=  F.C(0,0)+ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end PRidge;

  function LR_Peaks(R: RadPair; N: Positive) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (1,1);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(2,1) :=  ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end LR_Peaks;

  function BT_Peaks(R: RadPair; N: Positive) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (1,1);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(1,2) :=  ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end BT_Peaks;

  function MPeaks(R: RadPair; N: Positive) return Fourier2 is
    S: constant Scalar := Scal(Flt(41)/Flt(64));
    F: Fourier2;
  begin
    F.T.P := (1,1);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(1,2) :=  S;
    F.C(2,1) :=  S;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end MPeaks;

  function PPeaks(R: RadPair; N: Positive) return Fourier2 is
    S: constant Scalar := Scal(Flt(41)/Flt(64));
    F: Fourier2;
  begin
    F.T.P := (1,1);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(1,2) :=  S;
    F.C(2,1) := -S;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end PPeaks;

  function PMPeaks(R: RadPair; N: Positive) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (1,1);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(2,0) :=  ScalOne;
    F.C(0,2) :=  ScalOne;
    F.C(2,2) := -ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end PMPeaks;

  function LR_Holes(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2 := LR_Peaks(R,1);
  begin
    Neg(F);
    F.C(0,0) :=  F.C(0,0)+ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end LR_Holes;

  function BT_Holes(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2 := BT_Peaks(R,1);
  begin
    Neg(F);
    F.C(0,0) :=  F.C(0,0)+ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end BT_Holes;

  function MHoles(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2 := MPeaks(R,1);
  begin
    Neg(F);
    F.C(0,0) :=  F.C(0,0)+ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end MHoles;

  function PHoles(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2 := PPeaks(R,1);
  begin
    Neg(F);
    F.C(0,0) :=  F.C(0,0)+ScalOne;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end PHoles;

  function PMHoles(R: RadPair; N: Positive) return Fourier2 is
    F: Fourier2 := PMPeaks(R,N);
  begin
    Neg(F);
    F.C(0,0) := F.C(0,0)+ScalOne;
    return F;
  end PMHoles;

  function HVRidge(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (0,0);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(0,0) :=  Quarter*ScalThree;
    F.C(2,0) := -ScalQuarter;
    F.C(0,2) := -ScalQuarter;
    F.C(2,2) := -ScalQuarter;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end HVRidge;

  function EdgeRidge(R: RadPair; N: Natural) return Fourier2 is
    F: Fourier2;
  begin
    F.T.P := (0,0);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    F.C(0,0) :=  Quarter*ScalThree;
    F.C(2,0) :=  ScalQuarter;
    F.C(0,2) :=  ScalQuarter;
    F.C(2,2) := -ScalQuarter;
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end EdgeRidge;

  function Peaks8(R: RadPair; N: Positive) return Fourier2 is
    M: constant Flt := Flt(41)/Flt(32);
    F,G: Fourier2;
  begin
    F.T.P := (0,0);
    F.T.R := R;
    F.C := Zero_FCoeff;
    F.E := Zero_FErr;
    G := F;
    F.C(1,1) :=  Scal(M);
    G.C(0,2) :=  ScalOne;
    G.C(2,0) := -ScalOne;
    Mult(G,F);
    for K in 1 .. N loop
      F := F*F;
    end loop;
    return F;
  end Peaks8;

--------------- used with DiagReflect

  function Perturb1(R: RadPair; L: Scalar) return Fourier2 is
    W: Fourier2 := CornerPeaks(R,0);
  begin
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb1;

  function Perturb2(R: RadPair; L: Scalar) return Fourier2 is
    W: Fourier2 := CornerPeaks(R,1);
  begin
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb2;

  function Perturb3(R: RadPair; L: Scalar) return Fourier2 is
    W: Fourier2 := MValley(R,1);
  begin
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb3;

  function Perturb4(R: RadPair; L: Scalar) return Fourier2 is
--- here same as Perturb3
    W: Fourier2 := MValley(R,1);
  begin
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb4;

  function Perturb5(R: RadPair; L: Scalar) return Fourier2 is
    M: constant Flt := Flt(1)/Flt(8);
    W: Fourier2;
  begin
    W := MValley(R,1)+Scal(M)*EdgePeaks(R,1);
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb5;

  function Perturb6(R: RadPair; L: Scalar) return Fourier2 is
    W: Fourier2 := MidPeak(R,4);
  begin
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb6;

  function Perturb7(R: RadPair; L: Scalar) return Fourier2 is
    W: Fourier2;
  begin
    W := ScalHalf*MidHole(R,2)+MidPeak(R,2);
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb7;

  function Perturb8(R: RadPair; L: Scalar) return Fourier2 is
    W: Fourier2;
  begin
    W := EdgePeaks(R,1)+Scal(Three/Four)*MidPeak(R,3);
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb8;

  function Perturb9(R: RadPair; L: Scalar) return Fourier2 is
    M2: constant Scalar := ScalQuarter;
    M1: constant Scalar := ScalOne-M2;
    W: Fourier2;
  begin
    W := HVRidge(R,0);
    W := M1*W+M2*W*W;
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb9;

  function Perturb10(R: RadPair; L: Scalar) return Fourier2 is
    M: constant Flt := Flt(100);
    W: Fourier2;
  begin
    W := MValley(R,1)*(PValley(R,1)*HVValley(R,0));
    Mult(M*L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb10;

  function Perturb11(R: RadPair; L: Scalar) return Fourier2 is
    W: Fourier2 := Peaks8(R,1);
  begin
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end Perturb11;

  function Perturb12(R: RadPair; L: Scalar) return Fourier2 is
    L11: constant Flt := Flt(51)/Flt(64);
    W: Fourier2 := Perturb11(R,Scal(L11));
  begin
    AddProd(L,MidPeak(R,3),W);
    return W;
  end Perturb12;

  function Perturb13(R: RadPair; L: Scalar) return Fourier2 is
    L11: constant Flt := Flt(97)/Flt(128);
    W: Fourier2 := Perturb11(R,Scal(L11));
  begin
    AddProd(L,MidPeak(R,3),W);
    return W;
  end Perturb13;

----------------- used with YReflect

  function YPerturb1(R: RadPair; L: Scalar) return Fourier2 is
    W: Fourier2;
  begin
    W := LR_Peaks(R,1);
    Mult(L,W);
    W.C(0,0) := W.C(0,0)+ScalOne-L;
    return W;
  end YPerturb1;

---------------

  function LowSqrtNegLap(Last: Positive) return ScalDiag is
    M,N: Integer;
    D: ScalDiag(1 .. Last);
  begin
    for K in 1 .. Last loop
      Coord(K,M,N);
      D(K) := Sqrt(Scal(M*M+N*N));
    end loop;
    return D;
  end LowSqrtNegLap;

  procedure Sobolize(V: in out ScalVec) is
    D: constant ScalDiag := LowSqrtNegLap(V'Last);
  begin
    Mult(D,V);
  end Sobolize;

  procedure Sobolize(A: in out ScalMat) is
    D: constant ScalDiag := LowSqrtNegLap(A'Last(1));
  begin
    Mult(D,A);
    Mult(A,Inverse(D));
    Show("asymmetry ",Sup(Norm1(AntiSymm(A))),1);
  end Sobolize;

  procedure UnSobolize(V: in out ScalVec) is
    D: constant ScalDiag := LowSqrtNegLap(V'Last);
  begin
    Mult(Inverse(D),V);
  end UnSobolize;

  procedure UnSobolize(A: in out ScalMat) is
    D: constant ScalDiag := LowSqrtNegLap(A'Last(1));
  begin
    Mult(A,D);
    Mult(Inverse(D),A);
  end UnSobolize;

  procedure InvNegLaplace(F: in out Fourier2) is
    K: Integer;
    C: FCoeff renames F.C;
    E: FErr renames F.E;
  begin
    for M in Freq1 loop
      for N in Freq2 loop
        K := M*M+N*N;
        if K=0 then
          C(0,0) := ScalZero;
        else
          Mult(Inv(Scal(K)),C(M,N));
        end if;
      end loop;
    end loop;
    if not Trunc then
      for M in FreqE1 loop
        for N in FreqE2 loop
          K := M*M+N*N;
          if K>0 then
            MultUp(InvUp(Flt(K)),E(M,N));
          end if;
        end loop;
      end loop;
    end if;
  end InvNegLaplace;

 ----------

  procedure MapA(P: in ScalVec; F: in Fourier2; G: in out Fourier2) is
    W: constant Fourier2 := Perturb(F.T.R,P(1));
  begin
    G := W*(F*F*F);
    Symmetrize(G);
    InvNegLaplace(G);
  end MapA;

  procedure DMapA(H: in out Fourier2) is
  begin
    H := H*GSave;
    Symmetrize(H);
    InvNegLaplace(H);
  end DMapA;

  procedure DMapA(P: in ScalVec; F: in Fourier2; H: in out Fourier2) is
--- uses global Saved, LSave, FSave, GSave
  begin
    if not (Saved and then P(1)=LSave and then F=FSave) then
      GSave := ScalThree*Perturb(F.T.R,P(1))*(F*F);
      FSave := F;
      LSave := P(1);
      Saved := True;
    end if;
    DMapA(H);
  end DMapA;

-----------------------------------------

  procedure NumEigenData(DY: in DMap; P: in Param; F: in Fourier2; T: in out ScalSimDiag) is
    N: constant Integer := T.D'Last;
    A: ScalMat(1..N,1..N);
  begin
    Contr.NumDerMatrix(DY,P,F,A);
    Sobolize(A);
    Symmetrize(A);
    SV_Num.Eigen(A,ScalVec(T.D),T.L);
    Orthogonalize(T.L);
  end NumEigenData;

-----------------------------------------

  function Compo1(F,F0: Fourier2) return Scalar is
  begin
    return IProd(F,F0)/IProd(F0,F0);
  end Compo1;

  function Proj1(F,F0: Fourier2) return Fourier2 is
  begin
    return Compo1(F,F0)*F0;
  end Proj1;

  procedure BifMap(Y: in Map; V: ScalVec; F: in Fourier2; G: in out Fourier2) is
    La: constant Scalar := V(-1);
    Mu: constant Scalar := V(-2);
    P: constant ScalVec(1..2) := (La,Mu);
    U: Fourier2;
  begin
    Y(P,F,G);
    if V(0)=ScalOne then
      UnPack(V,F,U);
      G := (G-Proj1(G,U))+Mu*U;
    end if;
  end BifMap;

  procedure DBifMap(DY: in DMap; V: ScalVec; F: in Fourier2; H: in out Fourier2) is
    La: constant Scalar := V(-1);
    Mu: constant Scalar := V(-2);
    P: constant ScalVec(1..2) := (La,Mu);
    U: Fourier2;
  begin
    DY(P,F,H);
    if V(0)=ScalOne then
      UnPack(V,F,U);
      H := H-Proj1(H,U);
    end if;
  end DBifMap;

-----

  procedure BifMapA(V: ScalVec; F: in Fourier2; G: in out Fourier2) is
  begin
    BifMap(MapA'Access,V,F,G);
  end BifMapA;

  procedure DBifMapA(V: ScalVec; F: in Fourier2; H: in out Fourier2) is
  begin
    DBifMap(DMapA'Access,V,F,H);
  end DBifMapA;

begin

  if not Trunc then Rounding.Set_Rounding_Up; end if;

  UpdateNormFacs((Zero,Zero));

end Fouriers2;
