with Ada.Text_IO, Interfaces.C.Strings, Ints, Ints.IO;
use Ada.Text_IO, Interfaces.C.Strings, Ints, Ints.IO;

package body MPFR.Floats is

  procedure Initialize(R: in out MPFloat) is
--- not used during declarations that assign values
  begin
--    Put_Line("initializing");
    MPFR_Init2(R.Val,Precision);
  end Initialize;

  procedure Adjust(R: in out MPFloat) is
--- also adjusts precision
    Rt: MPFR_T := R.Val;
  begin
--    Put_Line("adjusting");
    MPFR_Init2(R.Val,Precision);
    MPFR_Set(R.Val,Rt,Rounding);
  end Adjust;

  procedure Finalize(R: in out MPFloat) is
--- also used during assignments, to finalize previous content
  begin
--   Put_Line("finalizing");
    MPFR_Clear(R.Val);
  end Finalize;

---------------

  procedure Set_Precision(Dummy: in MPFloat; Binary_Digits: Positive) is
  begin
    Precision := MP_Prec_T(Binary_Digits);
    MPFR_Set_Default_Prec(Precision);      -- not actually used
  end Set_Precision;

  function Get_Precision(R: MPFloat) return Positive is
  begin
    return Positive(MPFR_Get_Prec(R.Val));
  end Get_Precision;

  procedure Set_Rounding_Up(Dummy: in MPFloat) is
  begin
    Rounding := GMP_RNDU;
  end Set_Rounding_Up;

  procedure Set_Rounding_Down(Dummy: in MPFloat) is
  begin
    Rounding := GMP_RNDD;
  end Set_Rounding_Down;

  procedure Set_Rounding_Nearest(Dummy: in MPFloat) is
  begin
    Rounding := GMP_RNDN;
  end Set_Rounding_Nearest;

  procedure Set_Rounding_Truncate(Dummy: in MPFloat) is
  begin
    Rounding := GMP_RNDZ;
  end Set_Rounding_Truncate;

  function IsZero(R: MPFloat) return Boolean is
  begin
    return (Sign(R) = 0);
  end IsZero;

  procedure SetZero(R: in out MPFloat) is
  begin
    MPFR_Set_Si(R.Val,0,Rounding);
  end SetZero;

  procedure Swap(R1,R2: in out MPFloat) is
  begin
    MPFR_Swap(R1.Val,R2.Val);
  end Swap;

  function Compare(R1,R2: MPFloat) return Integer is
    I: constant Int := MPFR_Cmp(R1.Val,R2.Val);
  begin
    if Problem then Report("Compare"); end if;
    return Integer(I);
  end Compare;

  function "<"(R1,R2: MPFloat) return Boolean is
  begin
    return (Compare(R1,R2) < 0);
  end "<";

  function "<="(R1,R2: MPFloat) return Boolean is
  begin
    return (Compare(R1,R2) <= 0);
  end "<=";

  function "="(R1,R2: MPFloat) return Boolean is
  begin
    return (Compare(R1,R2) = 0);
  end "=";

  function ">="(R1,R2: MPFloat) return Boolean is
  begin
    return (Compare(R1,R2) >= 0);
  end ">=";

  function ">"(R1,R2: MPFloat) return Boolean is
  begin
    return (Compare(R1,R2) > 0);
  end ">";

  function Problem return Boolean is
    Found: Boolean := False;
  begin
    if MPFR_Underflow_P>0  then Put("Underflow "); Found := True; end if;
    if MPFR_Overflow_P>0   then Put("Overflow ");  Found := True; end if;
    if MPFR_NanFlag_P>0    then Put("NaN ");       Found := True; end if;
    if MPFR_ERangeFlag_P>0 then Put("ERange ");    Found := True; end if;
    return Found;
  end Problem;

  procedure Report(Name: in String) is
  begin
    Put_Line("error occured in " & Name);
    raise MPFR_Exception;
  end Report;

  function Sign(R: MPFloat) return Integer is
    I: constant Int := MPFR_Sgn(R.Val);
  begin
    if Problem then Report("Sign"); end if;
    return Integer(I);
  end Sign;

  function "abs"(R: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,MPFR_Get_Prec(R.Val));
    MPFR_Abs(Rt,R.Val,Rounding);
    if Problem then Report("abs"); end if;
    return (AFC with Rt);
  end "abs";

  function Min(R1,R2: MPFloat) return MPFloat is
    R: MPFR_T;
  begin
    if MPFR_Cmp(R1.Val,R2.Val)<0 then
      MPFR_Init2(R,MPFR_Get_Prec(R1.Val));
      MPFR_Set(R,R1.Val,Rounding);
    else
      MPFR_Init2(R,MPFR_Get_Prec(R2.Val));
      MPFR_Set(R,R2.Val,Rounding);
    end if;
    if Problem then Report("Min"); end if;
    return (AFC with R);
  end Min;

  function Max(R1,R2: MPFloat) return MPFloat is
    R: MPFR_T;
  begin
    if MPFR_Cmp(R1.Val,R2.Val)>0 then
      MPFR_Init2(R,MPFR_Get_Prec(R1.Val));
      MPFR_Set(R,R1.Val,Rounding);
    else
      MPFR_Init2(R,MPFR_Get_Prec(R2.Val));
      MPFR_Set(R,R2.Val,Rounding);
    end if;
    if Problem then Report("Max"); end if;
    return (AFC with R);
  end Max;

  function "-"(R: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,MPFR_Get_Prec(R.Val));
    MPFR_Neg(Rt,R.Val,Rounding);
    if Problem then Report("-"); end if;
    return (AFC with Rt);
  end "-";

  function "+"(R1,R2: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Add(Rt,R1.Val,R2.Val,Rounding);
    if Problem then Report("+"); end if;
    return (AFC with Rt);
  end "+";

  function "-"(R1,R2: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Sub(Rt,R1.Val,R2.Val,Rounding);
    if Problem then Report("+"); end if;
    return (AFC with Rt);
  end "-";

  function "*"(R1,R2: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Mul(Rt,R1.Val,R2.Val,Rounding);
    if Problem then Report("*"); end if;
    return (AFC with Rt);
  end "*";

  function "/"(R1,R2: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Div(Rt,R1.Val,R2.Val,Rounding);
    if Problem then Report("/"); end if;
    return (AFC with Rt);
  end "/";

  function "**"(R: MPFloat; N: Integer) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Pow_Si(Rt,R.Val,Long(N),Rounding);
    if Problem then Report("**"); end if;
    return (AFC with Rt);
  end "**";

  function "*"(F: Flt; R: MPFloat) return MPFloat is
    RF,Rt: MPFR_T;
  begin
    MPFR_Init2(RF,Long_Double'Machine_Mantissa);
    MPFR_Set_Ld(RF,Long_Double(F),Rounding);
    MPFR_Init2(Rt,Precision);
    MPFR_Mul(Rt,RF,R.Val,Rounding);
    if Problem then Report("("); end if;
    MPFR_Clear(RF);
    return (AFC with Rt);
  end "*";

  function Upper(R: MPFloat) return Flt is
  begin
    return Flt(MPFR_Get_Ld(R.Val,GMP_RNDU));
  end Upper;

  function Lower(R: MPFloat) return Flt is
  begin
    return Flt(MPFR_Get_Ld(R.Val,GMP_RNDD));
  end Lower;

  function Approx(R: MPFloat) return Flt is
  begin
    return Flt(MPFR_Get_Ld(R.Val,GMP_RNDN));
  end Approx;

  function ApproxPut(R: MPFloat) return FltVector is
    NR: constant MP_Prec_T := MPFR_Get_Prec(R.Val);
    NV: constant MP_Prec_T := MP_Prec_T(Long_Double'Machine_Mantissa);
    V: FltVector(1 .. Integer(NR)/Integer(NV)+1);
    R1,R2,VN: MPFR_T;
  begin
    MPFR_Init2(R1,NR);
    MPFR_Init2(R2,NR);
    MPFR_Init2(VN,NV+1);
    MPFR_Set(R1,R.Val,GMP_RNDZ);
    for N in V'Range loop
      MPFR_Set(R2,R1,GMP_RNDZ);
      V(N) := Flt(MPFR_Get_Ld(R2,GMP_RNDZ));
      MPFR_Set_Ld(VN,Long_Double(V(N)),GMP_RNDZ);
      MPFR_Sub(R1,R2,VN,GMP_RNDZ);
    end loop;
    MPFR_Clear(VN);
    MPFR_Clear(R2);
    MPFR_Clear(R1);
    return V;
  end ApproxPut;

  function ApproxGet(V: FltVector) return MPFloat is
    NV: constant MP_Prec_T := MP_Prec_T(Long_Double'Machine_Mantissa);
    R1,R2,VN: MPFR_T;
  begin
    MPFR_Init2(R1,Precision);
    MPFR_Init2(R2,Precision);
    MPFR_Init2(VN,NV+1);
    MPFR_Set_Si(R1,0,GMP_RNDZ);
    for N in reverse V'Range loop
      MPFR_Set(R2,R1,GMP_RNDZ);
      MPFR_Set_Ld(VN,Long_Double(V(N)),GMP_RNDZ);
      MPFR_Add(R1,R2,VN,GMP_RNDZ);
    end loop;
    MPFR_Clear(VN);
    MPFR_Clear(R2);
    return (AFC with R1);
  end ApproxGet;

  function Val(I: Integer) return MPFloat is
    RI: MPFR_T;
  begin
    MPFR_Init2(RI,Precision);
    MPFR_Set_Si(RI,Long(I),Rounding);
    return (AFC with RI);
  end Val;

  function Val(F: Flt) return MPFloat is
    RF: MPFR_T;
  begin
    MPFR_Init2(RF,Precision);
    MPFR_Set_Ld(RF,Long_Double(F),Rounding);
    return (AFC with RF);
  end Val;

---------------

  function Sqrt(R: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Sqrt(Rt,R.Val,Rounding);
    if Problem then Report("Sqrt"); end if;
    return (AFC with Rt);
  end Sqrt;

  function Exp(R: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Exp(Rt,R.Val,Rounding);
    if Problem then Report("Exp"); end if;
    return (AFC with Rt);
  end Exp;

  function Log(R: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Log(Rt,R.Val,Rounding);
    if Problem then Report("Log"); end if;
    return (AFC with Rt);
  end Log;

  function Cos(R: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Cos(Rt,R.Val,Rounding);
    if Problem then Report("Cos"); end if;
    return (AFC with Rt);
  end Cos;

  function Sin(R: MPFloat) return MPFloat is
    Rt: MPFR_T;
  begin
    MPFR_Init2(Rt,Precision);
    MPFR_Sin(Rt,R.Val,Rounding);
    if Problem then Report("Sin"); end if;
    return (AFC with Rt);
  end Sin;

  procedure Clear_Flags is
  begin
    if Problem then Report("Clear_Flags"); end if;
    MPFR_Clear_Flags;
  end Clear_Flags;

  function Raised_Inexact return Boolean is
    I: constant Int := MPFR_InexFlag_P;
  begin
    Clear_Flags;
    return (I>0);
  end Raised_Inexact;

  function NDigits(P: MP_Prec_T; Base: Positive) return Size_T is
    N: constant Positive := 1073741823; -- 2**30-1
  begin
    return Size_T((Width(N,Base)/30)*Positive(P));
  end NDigits;

  function EffLast(F: Positive; S: String) return Positive is
  begin
    for N in reverse F+1 .. S'Last loop
      if S(N) /= '0' then return N; end if;
    end loop;
    return F;
  end EffLast;

  function Image(R: MPFloat; Base: Positive; Digs: Natural := 0) return String is
    B: constant MPFR_Base := MPFR_Base(Base);
    EP: aliased MP_Exp_T_Ptr := new MP_Exp_T'(0);
    C: Chars_Ptr;
    D: Size_T := Size_T(Digs);
  begin
    if IsZero(R) then Just_Rounded := False; return " 0.0+0"; end if;
    if Digs=0 and then Strip0(Base,2) /= 1 then
      D := NDigits(MPFR_Get_Prec(R.Val),Base); -- full precision impossible
    end if;
    Clear_Flags;
    C := MPFR_Get_Str(Null_Ptr,EP,B,0,R.Val,Rounding);
    Just_Rounded := Raised_Inexact;
    declare
      E: constant Integer := Integer(EP.all-1);
      M: constant String := Value(C);
    begin
      MPFR_Free_Str(C);
      if Sign(R)<0 then
        return '-' & M(2) & '.' & M(3 .. EffLast(3,M)) & Image(E,Base,True);
      else
        return '+' & M(1) & '.' & M(2 .. EffLast(2,M)) & Image(E,Base,True);
      end if;
    end;
  end Image;

  function Value(S: String; Base: Positive) return MPFloat is
    B: constant MPFR_Base := MPFR_Base(Base);
    NoDot: Boolean := True;
    E: Integer;
    C: Chars_Ptr;
    R: MPFR_T;
  begin
    for K in S'Range loop
      if NoDot then
        NoDot := (S(K) /= '.');
      elsif S(K)='+' or else S(K)='-' then
        E := Value(S(K .. S'Last),Base);
        C := New_String(S(S'First .. K-1) & "@" & Image(E,10,True));
        exit;
      elsif K=S'Last then
        C := New_String(S & "@+0");
      end if;
    end loop;
    Clear_Flags;
    MPFR_Init_Set_Str(R,C,B,Rounding);
    Just_Rounded := Raised_Inexact;
    Free(C);
    return (AFC with R);
  end Value;

  function ValDec(S: String) return MPFloat is
  begin
    return Value(S,10);
  end ValDec;

  function ValHex(S: String) return MPFloat is
  begin
    return Value(S,16);
  end ValHex;

  function DecStr(R: MPFloat; Digs: Natural := 0) return String is
  begin
    return Image(R,10,Digs);
  end DecStr;

  function HexStr(R: MPFloat; Digs: Natural := 0) return String is
  begin
    return Image(R,16,Digs);
  end HexStr;

  function Rounded_Last_Conversion return Boolean is
  begin
    return Just_Rounded;
  end Rounded_Last_Conversion;

begin

  if Long_Double'Machine_Mantissa /= Flt'Machine_Mantissa then
    raise MPFR_Exception with "adapt precision of Flt";
  end if;

  if Long'Size < Integer'Size then
    raise MPFR_Exception with "size mismatch";
  end if;

end MPFR.Floats;
