with Ints.IO, Messages;
use Ints.IO, Messages;

package body Vectors.Ops is

  function Norm1(V: Vector) return Scalar is
    S: Scalar;
  begin
    SetZero(S);
    for I in V'Range loop
      S := S+abs(V(I));
    end loop;
    return S;
  end Norm1;

  function Norm1(A: Matrix) return Scalar is
    S,T: Scalar;
  begin
    SetZero(T);
    for J in A'Range(2) loop
      SetZero(S);
      for I in A'Range(1) loop
        S := S+abs(A(I,J));
      end loop;
      T := Max(S,T);
    end loop;
    return T;
  end Norm1;

  procedure Normalize2(V: in out Vector) is
    S: constant Scalar := Inv(Sqrt(V*V));
  begin
    Mult(S,V);
  end Normalize2;

  procedure Identity(A: in out Matrix) is
  begin
    for I in A'Range(2) loop
      for J in A'Range(1) loop SetZero(A(I,J)); end loop;
      A(I,I) := Scal(1);
    end loop;
  end Identity;

  function Identity(N: Positive) return Matrix is
    A: Matrix(1..N,1..N);
  begin
    Identity(A);
    return A;
  end Identity;

  procedure Symmetrize(A: in out Matrix) is
    SHalf: constant Scalar := Scal(1)/Scal(2);
  begin
    for I in A'Range(1) loop
      for J in A'First(2) .. I-1 loop
        A(I,J) := SHalf*(A(I,J)+A(J,I));
        A(J,I) := A(I,J);
      end loop;
    end loop;
  end Symmetrize;

  procedure CheckSymm(A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      for J in A'First(2) .. I-1 loop
        A(I,J) := Intersection(A(I,J),A(J,I));
        A(J,I) := A(I,J);
      end loop;
    end loop;
  end CheckSymm;

  function AntiSymm(A: in Matrix) return Matrix is
    ScalHalf: constant Scalar := Scal(1)/Scal(2);
    B: Matrix(A'Range(1),A'Range(2));
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        B(I,J) := ScalHalf*(A(I,J)-A(J,I));
      end loop;
      SetZero(B(I,I));
    end loop;
    return B;
  end AntiSymm;

  procedure Invert(D: in out Diag) is
  begin
    for I in D'Range loop
      D(I) := Inv(D(I));
    end loop;
  end Invert;

  function Inverse(D: in Diag) return Diag is
    Dt: Diag(D'Range);
  begin
    for I in D'Range loop
      Dt(I) := Inv(D(I));
    end loop;
    return Dt;
  end Inverse;

  procedure LTriangSolve(A: in Matrix; V: in out Vector) is
    N: constant Integer := V'Last;
    S: Scalar;
  begin
    for I in 1 .. N loop
      S := V(I);
      for J in 1 .. I-1 loop
        S := S-A(I,J)*V(J);
      end loop;
      V(I) := S/A(I,I);
    end loop;
  end LTriangSolve;

  procedure UTriangSolve(A: in Matrix; V: in out Vector) is
    N: constant Integer := V'Last;
    S: Scalar;
  begin
    for I in reverse 1 .. N loop
      S := V(I);
      for J in I+1 .. N loop
        S := S-A(I,J)*V(J);
      end loop;
      V(I) := S/A(I,I);
    end loop;
  end UTriangSolve;

  procedure LTriangInv(A: in Matrix; B: in Matrix; X: in out Matrix) is
    N: constant Integer := X'Last(1);
    S: Scalar;
  begin
    TraceEnter("Vectors.Ops.LTriangInv");
    for K in 1 .. N loop
      for I in 1 .. N loop
        S := B(I,K);
        for J in 1 .. I-1 loop
          S := S-A(I,J)*X(J,K);
        end loop;
        X(I,K) := S/A(I,I);
      end loop;
    end loop;
    TraceLeave;
  end LTriangInv;

  procedure UTriangInv(A: in Matrix; B: in Matrix; X: in out Matrix) is
    N: constant Integer := X'Last(1);
    S: Scalar;
  begin
    TraceEnter("Vectors.Ops.UTriangInv");
    for K in 1 .. N loop
      for I in reverse 1 .. N loop
        S := B(I,K);
        for J in I+1 .. N loop
          S := S-A(I,J)*X(J,K);
        end loop;
        X(I,K) := S/A(I,I);
      end loop;
    end loop;
    TraceLeave;
  end UTriangInv;

  procedure FindMax(A: in Matrix; K: in Integer; II,JJ: out Integer) is
    R,R0: Flt := Zero;
  begin
    II := K;
    JJ := K;
    for I in K .. A'Last(1) loop
      for J in K .. A'Last(2) loop
        R := MaxNorm(A(I,J));
        if R>R0 then R0 := R; II := I; JJ := J; end if;
      end loop;
    end loop;
  end FindMax;

  procedure Orthogonalize(A: in out Matrix) is
    type Column is new Vector(A'Range(1));
    VJ: Column;
    F: Vector(A'Range(2));
    U: array(A'Range(2)) of Column;
  begin
    for J in F'Range loop
      GetColumn(J,A,VJ);
      U(J) := VJ;
      F(J) := Inv(VJ*VJ);
      for K in F'First .. J-1 loop
        AddProd(-(VJ*U(K))*F(K),U(K),U(J));
      end loop;
    end loop;
    for J in F'Range loop
      Mult(Sqrt(F(J)),U(J));
      SetColumn(J,U(J),A);
    end loop;
  end Orthogonalize;

  function OrthoErr(A: in Matrix) return Matrix is
    B: Matrix := Transpose(A)*A;
  begin
    for K in B'Range(1) loop
      SetZero(B(K,K));
    end loop;
    return B;
  end OrthoErr;

  procedure Decomp(A: in Matrix; T: in out Cholesky) is
    N: constant Integer := A'First(1);
    S: Scalar;
    D: Diag renames T.D;
    L: Matrix renames T.L;
  begin
    Identity(L);
    D(1) := A(1,1);
    for I in 2 .. N loop
      for J in 1 .. I-1 loop
        S := A(I,J);
        for K in 1 .. J-1 loop
          S := S-L(I,K)*L(J,K)*D(K);
        end loop;
        L(I,J) := S/D(J);
      end loop;
      S := A(I,I);
      for K in 1 .. I-1 loop
        S := S-L(I,K)*L(I,K)*D(K);
      end loop;
      D(I) := S;
    end loop;
  end Decomp;

  procedure Decomp(A: in Matrix; T: in out LUData) is
    P: Pivot renames T.P;
    Q: Pivot renames T.Q;
    InvL: Matrix renames T.InvL;
    U: Matrix renames T.U;
    N: constant Positive := P'Last;
    UKK,S: Scalar;
  begin
    SetZero(InvL);
    U := A;
    for K in 1 .. N loop
      FindMax(U,K,P(K),Q(K));
      SwapRows(K,P(K),InvL);
      SwapRows(K,P(K),U);
      SwapColumns(K,Q(K),U);
      InvL(K,K) := Scal(1);
      UKK := U(K,K);
      for I in K+1 .. N loop
        S := U(I,K)/UKK;
        for J in 1 .. K loop
          InvL(I,J) := InvL(I,J)-S*InvL(K,J);
        end  loop;
        SetZero(U(I,K));
        for J in K+1 .. N loop
          U(I,J) := U(I,J)-S*U(K,J);
        end loop;
      end loop;
    end loop;
  end Decomp;

  function Decomp(A: Matrix) return LUData is
    T: LUData(A'Last(1));
  begin
    Decomp(A,T);
    return T;
  end Decomp;

  procedure Permute(P: in Pivot; V: in out Vector) is
    K: Integer;
  begin
    for I in reverse P'Range loop
      K := P(I);
      if K /= I then
        Swap(V(I),V(K));
      end if;
    end loop;
  end Permute;

  procedure UnPermute(P: in Pivot; V: in out Vector) is
    K: Integer;
  begin
    for I in P'Range loop
      K := P(I);
      if K /= I then
        Swap(V(I),V(K));
      end if;
    end loop;
  end UnPermute;

  procedure Mult(T: in LUData; V: in out Vector) is
  begin
    UnPermute(T.Q,V);
    Mult(T.U,V);
    LTriangSolve(T.InvL,V);
    Permute(T.P,V);
  end Mult;

  procedure Solve(T: in LUData; V: in out Vector) is
  begin
    UnPermute(T.P,V);
    Mult(T.InvL,V);
    UTriangSolve(T.U,V);
    Permute(T.Q,V);
  end Solve;

  procedure LDLTransposed(D: in Diag; L: in Matrix ; A: in out Matrix) is
  begin
    A := L;
    Transpose(A);
    Mult(D,A);
    Mult(L,A);
  end LDLTransposed;

  procedure Convert(T: in DL; A: in out Matrix) is
  begin
    LDLTransposed(T.D,T.L,A);
  end Convert;

  procedure Convert(T: in LUData; A: in out Matrix) is
  begin
    LTriangInv(T.InvL,T.U,A);
    for K in reverse T.P'Range loop
      SwapRows(K,T.P(K),A);
      SwapColumns(K,T.Q(K),A);
    end loop;
  end Convert;

  procedure Invert(T: in Cholesky; A: in out Matrix) is
    N: constant Integer := T.D'Last;
  begin
    LTriangInv(T.L,Identity(N),A);
    A := Transpose(A)*(Inverse(T.D)*A);
  end Invert;

  procedure Invert(T: in LUData; A: in out Matrix) is
  begin
    UTriangInv(T.U,T.InvL,A);
    for K in reverse T.P'Range loop
      SwapRows(K,T.Q(K),A);
      SwapColumns(K,T.P(K),A);
    end loop;
  end Invert;

  procedure Solve(A: in Matrix; B: in Vector; X: in out Vector) is
  begin
    X := B;
    Solve(Decomp(A),X);
  end Solve;

  procedure Invert(A: in out Matrix) is
  begin
    Invert(Decomp(A),A);
  end Invert;

  procedure InvertSym(A: in out Matrix) is
    T: Cholesky(A'Last(1));
  begin
    Decomp(A,T);
    Invert(T,A);
  end InvertSym;

  procedure Write(Name: in String; T: in DL; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Writing " & Name);
    Create(F,Out_File,Name);
    Put(F,T.D,Decimal);
    Put(F,T.L,Decimal);
    Close(F);
  end Write;

  procedure Write(Name: in String; T: in LUData; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    Put_Line("Writing " & Name);
    Create(F,Out_File,Name);
    Put(F,T.P);
    Put(F,T.Q);
    Put(F,T.InvL,Decimal);
    Put(F,T.U,Decimal);
    Close(F);
  end Write;

  function Read(Name: String; Decimal: Boolean := False) return DL is
    F: File_Type;
  begin
    Put_Line("Reading " & Name);
    Open(F,In_File,Name);
    declare
      D: Diag := Get(F,1);
      T: DL(D'Last);
    begin
      T.D := D;
      Get(F,T.L,Decimal);
      Close(F);
      return T;
    end;
  end Read;

  function Read(Name: String; Decimal: Boolean := False) return LUData is
    F: File_Type;
  begin
    Put_Line("Reading " & Name);
    Open(F,In_File,Name);
    declare
      P: Pivot := Get(F,1);
      T: LUData(P'Last);
    begin
      T.P := P;
      T.Q := Get(F,1);
      Get(F,T.InvL,Decimal);
      Get(F,T.U,Decimal);
      Close(F);
      return T;
    end;
  end Read;

end Vectors.Ops;
